# Copyright 2017 Hynek Schlawack

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
import os

import attr

from .exceptions import MissingEnvValueError


CNF_KEY = "environ_config"
log = logging.getLogger(CNF_KEY)


# We define a sentinel for when prefixes are not set and special handling for
# what value to use at the app level when that is the case
class Sentinel:
    def __init__(self, bool_=True):
        self._bool = bool(bool_)

    def __bool__(self):
        return self._bool


PREFIX_NOT_SET = Sentinel(False)
DEFAULT_PREFIX = "APP"


def _get_prefix(obj):
    """
    Get the prefix of an environ object or the default value if it is not set.
    """
    return obj._prefix if obj._prefix is not PREFIX_NOT_SET else DEFAULT_PREFIX


@attr.s
class Raise:
    pass


RAISE = Raise()


def config(
    maybe_cls=None,
    prefix=PREFIX_NOT_SET,
    from_environ="from_environ",
    generate_help="generate_help",
    frozen=False,
):
    """
    Make a class a configuration class.

    :param str prefix: The prefix that is used for the env variables.  If you
        have an `var` attribute on the class and you leave the default argument
        value of *PREFIX_NOT_SET*, the *DEFAULT_PREFIX* value of ``APP`` will
        be used and *environ-config* will look for an environment variable
        called ``APP_VAR``.
    :param str from_environ: If not `None`, attach a config loading method with
        the name *from_environ* to the class.  See `to_config` for more
        information.
    :param str generate_help: If not `None`, attach a config loading method
        with the name *generate_help* to the class.  See `generate_help` for
        more information.
    :param bool frozen: The configuration will be immutable after
        instantiation, if `True`.

    .. versionadded:: 19.1.0
       *from_environ*
    .. versionadded:: 19.1.0
       *generate_help*
    .. versionadded:: 20.1.0
       *frozen*
    .. versionchanged:: 21.1.0
       *prefix* now defaults to *PREFIX_NOT_SET* instead of ``APP``.
    """

    def wrap(cls):
        def from_environ_fnc(cls, environ=os.environ):
            return __to_config(cls, environ)

        def generate_help_fnc(cls, **kwargs):
            return __generate_help(cls, **kwargs)

        cls._prefix = prefix
        if from_environ is not None:
            setattr(cls, from_environ, classmethod(from_environ_fnc))
        if generate_help is not None:
            setattr(cls, generate_help, classmethod(generate_help_fnc))
        return attr.s(cls, frozen=frozen, slots=True)

    if maybe_cls is None:
        return wrap
    else:
        return wrap(maybe_cls)


@attr.s(slots=True)
class _ConfigEntry:
    name = attr.ib(default=None)
    default = attr.ib(default=RAISE)
    sub_cls = attr.ib(default=None)
    callback = attr.ib(default=None)
    help = attr.ib(default=None)


def var(default=RAISE, converter=None, name=None, validator=None, help=None):
    """
    Declare a configuration attribute on the body of `config`-decorated class.

    It will be attempted to be filled from an environment variable based on the
    prefix and *name*.

    :param default: Setting this to a value makes the config attribute
        optional.
    :param str name: Overwrite name detection with a string.  If not set, the
        name of the attribute is used.
    :param converter: A callable that is run with the found value and
        its return value is used.  Please not that it is also run for default
        values.
    :param validator: A callable that is run with the final value.
        See ``attrs``'s `chapter on validation
        <https://www.attrs.org/en/stable/init.html#validators>`_ for details.
        You can also use any validator that is `shipped with attrs
        <https://www.attrs.org/en/stable/api.html#validators>`_.
    :param str help: A help string that is used by `generate_help`.
    """
    return attr.ib(
        default=default,
        metadata={CNF_KEY: _ConfigEntry(name, default, None, None, help)},
        converter=converter,
        validator=validator,
    )


def _env_to_bool(val):
    """
    Convert *val* to a bool if it's not a bool in the first place.
    """
    if isinstance(val, bool):
        return val
    val = val.strip().lower()
    if val in ("1", "true", "yes"):
        return True

    return False


def bool_var(default=RAISE, name=None, help=None):
    """
    Like `var`, but converts the value to a `bool`.

    The following values are considered `True`:

    - ``True`` (if you set a *default*)
    - ``"1"``
    - ``"true"``
    - ``"yes"``

    Every other value is interpreted as `False`.  Leading and trailing
    whitespace is ignored.
    """
    return var(default=default, name=name, converter=_env_to_bool, help=help)


def group(cls, optional=False):
    """
    A configuration attribute that is another configuration class.

    This way you can nest your configuration hierarchically although the values
    are coming from a flat source.

    The group's name is used to build a namespace::

       @environ.config
       class Cfg:
           @environ.config
           class Sub:
               x = environ.var()

           sub = environ.group(Sub)

    The value of ``x`` is looked up using ``APP_SUB_X``.

    You can nest your configuration as deeply as you wish.

    The *optional* keyword argument can be used to mark a *group* referencing
    a "child" *config* object so that if all variables defined in the child
    (including sub-groups) are not present in the environment being parsed, the
    attribute corresponding to the *optional* *group* will be set to `None`.

    :param bool optional: Mark this group as *optional*. Defaults to `False`.

    :returns: An attribute which will be used as a nested *group* of variables.

    .. versionadded:: 21.1.0
       *optional*
    """
    default = None if optional else RAISE
    return attr.ib(
        default=default,
        metadata={CNF_KEY: _ConfigEntry(None, default, cls, True)},
    )


def _default_getter(environ, metadata, prefix, name):
    """
    This default lookup implementation simply gets values from *environ*.
    """
    ce = metadata[CNF_KEY]
    if ce.name is not None:
        var = ce.name
    else:
        var = "_".join(prefix + (name,)).upper()
    log.debug("looking for env var '%s'.", var)
    try:
        return environ[var]
    except KeyError:
        raise MissingEnvValueError(var)


def _to_config_recurse(config_cls, environ, prefixes, default=RAISE):
    """
    Traverse *config_cls* to construct an instance with values from *environ*.

    This function walks through a potential tree of config definition classes
    and uses the specified (via attributes set through class construction) or
    default implementation of config variable lookup to collect values from the
    provided *environ* object. The collected configuration values (including
    sub-config objects, e.g. for groups) are used to instantiate the
    well-structured *config_cls* with those values being accessible via the new
    object's attributes.
    """
    # We keep track of values we actually got from the getter vs those we set
    # from the `ConfigEntry` default value
    got = {}
    defaulted = {}
    missing_vars = set()

    for attr_obj in attr.fields(config_cls):
        try:
            ce = attr_obj.metadata[CNF_KEY]
        except KeyError:
            continue
        name = attr_obj.name

        if ce.sub_cls is not None:
            prefix = ce.sub_cls._prefix or name
            got[name] = _to_config_recurse(
                ce.sub_cls, environ, prefixes + (prefix,), default=ce.default
            )
        else:
            getter = ce.callback or _default_getter
            try:
                got[name] = getter(environ, attr_obj.metadata, prefixes, name)
            except MissingEnvValueError as exc:
                if isinstance(ce.default, Raise):
                    missing_vars |= set(exc.args)
                else:
                    defaulted[name] = (
                        attr.NOTHING
                        if isinstance(ce.default, attr.Factory)
                        else ce.default
                    )

    if missing_vars:
        # If we were told to raise OR if we got *any* values for our attrs, we
        # will raise a `MissingEnvValueError` with all the missing variables
        if isinstance(default, Raise) or got:
            raise MissingEnvValueError(*missing_vars)
        # otherwise we will simply use the default passed into this call
        else:
            # Should be no need to handle `Factory`s here
            return default
    # Merge the defaulted and actually collected values into the config type
    defaulted.update(got)
    return config_cls(**defaulted)


def to_config(config_cls, environ=os.environ):
    """
    Load the configuration as declared by *config_cls* from *environ*.

    :param config_cls: The configuration class to fill.
    :param dict environ: Source of the configuration.  `os.environ` by default.

    :returns: An instance of *config_cls*.

    This is equivalent to calling ``config_cls.from_environ()``.
    """
    # The canonical app prefix might be falsey in which case we'll still set
    # the default prefix for this top level config object
    app_prefix = tuple(p for p in (_get_prefix(config_cls),) if p)
    return _to_config_recurse(config_cls, environ, app_prefix)


def _format_help_dicts(help_dicts, display_defaults=False):
    """
    Format the output of _generate_help_dicts into a str
    """
    help_strs = []
    for help_dict in help_dicts:
        help_str = "{} ({}".format(
            help_dict["var_name"],
            "Required" if help_dict["required"] else "Optional",
        )
        if help_dict.get("default") and display_defaults:
            help_str += ", Default=%s)" % help_dict["default"]
        else:
            help_str += ")"
        if help_dict.get("help_str"):
            help_str += ": %s" % help_dict["help_str"]
        help_strs.append(help_str)

    return "\n".join(help_strs)


def _generate_var_name(prefix, field_name):
    """
    Generate the environment variable name, given a prefix
    and the configuration field name.

    Examples:

    >>> _generate_var_name("", "some_var")
    "SOME_VAR"
    >>> _generate_var_name("my_app", "some_var")
    "MY_APP_SOME_VAR"

    :param prefix: the prefix to be used, can be empty
    :param field_name: the name of the field from which the variable is derived
    """
    return (
        "_".join((prefix, field_name)).upper()
        if prefix
        else field_name.upper()
    )


def _generate_new_prefix(current_prefix, class_name):
    """
    Generate the new prefix to be used when handling nested configurations.

    Examples:

    >>> _generate_new_prefix("", "config_group_1")
    "CONFIG_GROUP_1"
    >>> _generate_new_prefix("my_app", "another_config_group")
    "MY_APP_ANOTHER_CONFIG_GROUP"

    :param prefix: the prefix to be used, can be empty
    :param field_name: the name of the field from which the variable is derived
    """
    return (
        "_".join((current_prefix, class_name)).upper()
        if current_prefix
        else class_name.upper()
    )


def _generate_help_dicts(config_cls, _prefix=PREFIX_NOT_SET):
    """
    Generate dictionaries for use in building help strings.

    Every dictionary includes the keys...

    var_name: The env var that should be set to populate the value.
    required: A bool, True if the var is required, False if it's optional.

    Conditionally, the following are included...

    default: Included if an optional variable has a default set
    help_str: Included if the var uses the help kwarg to provide additional
        context for the value.

    Conditional key inclusion is meant to differentiate between exclusion
    vs explicitly setting a value to None.
    """
    help_dicts = []
    _prefix = _prefix if _prefix is not PREFIX_NOT_SET else config_cls._prefix
    for a in attr.fields(config_cls):
        try:
            ce = a.metadata[CNF_KEY]
        except KeyError:
            continue
        if ce.sub_cls is None:  # Base case for "leaves".
            if ce.name is None:
                var_name = _generate_var_name(_prefix, a.name)
            else:
                var_name = ce.name
            req = isinstance(ce.default, Raise)
            help_dict = {"var_name": var_name, "required": req}
            if not req:
                help_dict["default"] = ce.default
            if ce.help is not None:
                help_dict["help_str"] = ce.help
            help_dicts.append(help_dict)
        else:  # Construct the new prefix and recurse.
            help_dicts += _generate_help_dicts(
                ce.sub_cls, _prefix=_generate_new_prefix(_prefix, a.name)
            )
    return help_dicts


def generate_help(config_cls, formatter=None, **kwargs):
    """
    Autogenerate a help string for a config class.

    :param formatter: A callable that will be called with the help dictionaries
       as an argument and the remaining *kwargs*.  It should return the help
       string.
    :param bool display_defaults: When using the default formatter, passing
       `True` for *display_defaults* makes the default values part of the
       output.

    :returns: A help string that can be printed to the user.

    This is equivalent to calling ``config_cls.generate_help()``.

    .. versionadded:: 19.1.0
    """
    if formatter is None:
        formatter = _format_help_dicts
    help_dicts = _generate_help_dicts(config_cls)

    return formatter(help_dicts, **kwargs)


# We need these aliases because of a name clash with a function argument.
__generate_help = generate_help
__to_config = to_config



# Copyright 2017 Hynek Schlawack

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import attr
import pytest

from attr.exceptions import FrozenInstanceError

import environ


@environ.config(prefix="XYZ")
class Nested:
    """
    A nested configuration example.
    """

    @environ.config
    class Sub:
        y = environ.var()

    x = environ.var()
    sub = environ.group(Sub)


@environ.config(prefix="FOO")
class Parent:
    not_a_var = attr.ib()  # For testing that only environ.var's are processed.
    var1 = environ.var(help="var1, no default")
    var2 = environ.var("bar", help="var2, has default")
    var3 = environ.bool_var(help="var3, bool_var, no default")
    var4 = environ.bool_var(True, help="var4, bool_var, has default")
    var5 = environ.var("canine", name="DOG", help="var5, named, has default")
    var6 = environ.var(name="CAT", help="var6, named, no default")

    @environ.config
    class Child:
        var7 = environ.var(help="var7, no default")
        var8 = environ.var("bar", help="var8, has default")
        var9 = environ.bool_var(help="var9, bool_var, no default")
        var10 = environ.bool_var(True, help="var10, bool_var, has default")
        var11 = environ.var(
            "canine", name="DOG2", help="var11, named, has default"
        )
        var12 = environ.var(name="CAT2", help="var12, named, no default")
        var13 = environ.var("default")  # var with default, no help
        var14 = environ.var()  # var without default, no help

    child = environ.group(Child)


@environ.config(prefix="")
class NoPrefix:
    a_var = environ.var(help="a_var, no default")
    another_var = environ.var("bar", help="another_var, has default")
    _start_with_underscore = environ.var(help="this starts with an underscore")


class TestEnvironConfig:
    def test_empty(self):
        """
        Empty config is accepted.
        """

        @environ.config
        class Empty:
            pass

        cfg = environ.to_config(Empty)

        assert "Empty()" == repr(cfg)

    def test_flat(self):
        """
        Flat config is extracted.
        """

        @environ.config(prefix="APP")
        class Flat:
            x = environ.var()
            y = environ.var()

        cfg = environ.to_config(Flat, environ={"APP_X": "foo", "APP_Y": "bar"})

        assert Flat(x="foo", y="bar") == cfg

    def test_nested(self):
        """
        Nested config is extracted, prefix and vault_prefix are propagated.
        """
        env = {"APP_X": "nope", "XYZ_X": "foo", "XYZ_SUB_Y": "bar"}
        cfg = environ.to_config(Nested, environ=env)

        assert Nested(x="foo", sub=Nested.Sub(y="bar")) == cfg

    def test_missing(self):
        """
        If a var is missing, a human-readable MissingEnvValueError is raised.
        """

        @environ.config
        class Mandatory:
            x = environ.var()

        with pytest.raises(environ.MissingEnvValueError) as e:
            environ.to_config(Mandatory, environ={"y": "boring"})

        assert ("APP_X",) == e.value.args

    def test_default(self):
        """
        Default values are used iff the vars are missing.
        """

        @environ.config
        class Defaults:
            x = environ.var("foo")
            y = environ.var("qux")

        cfg = environ.to_config(Defaults, environ={"APP_Y": "bar"})

        assert Defaults(x="foo", y="bar") == cfg

    def test_factory_default(self):
        """
        If the default value is an ``attr.Factory``,
        it used to generate the default value.
        """

        @environ.config
        class Defaults:
            x = environ.var(attr.Factory(list))
            y = environ.var(attr.Factory(list))

        cfg = environ.to_config(Defaults, environ={"APP_Y": "bar"})

        assert Defaults(x=[], y="bar") == cfg

    @pytest.mark.parametrize("prefix", [None, ""])
    def test_no_prefix(self, prefix):
        """
        If prefix is None or "", don't add a leading _ when adding namespaces.
        """

        @environ.config(prefix=prefix)
        class Cfg:
            x = environ.var()

        cfg = environ.to_config(Cfg, environ={"X": "foo"})

        assert Cfg("foo") == cfg

    def test_overwrite(self):
        """
        The env variable name can be overwritten.
        """

        @environ.config
        class Cfg:
            x = environ.var(name="LANG")
            y = environ.var()

        cfg = environ.to_config(
            Cfg, environ={"APP_X": "nope", "LANG": "foo", "APP_Y": "bar"}
        )

        assert Cfg("foo", "bar") == cfg

    def test_no_prefixes(self):
        """
        If no prefixes are wished, nothing is prepended.
        """

        @environ.config(prefix=None)
        class Cfg:
            @environ.config
            class Sub:
                y = environ.var()

            x = environ.var()
            y = environ.var()
            sub = environ.group(Sub)

        cfg = environ.to_config(
            Cfg, environ={"X": "x", "Y": "y", "SUB_Y": "sub_y"}
        )

        assert Cfg("x", "y", Cfg.Sub("sub_y")) == cfg

    @pytest.mark.parametrize("val", ["1", "trUe", "yEs  "])
    def test_bool_var(self, val):
        """
        Truthy strings are converted to True, everything else to False.

        Defaults can be passed as bools.
        """

        @environ.config
        class Cfg:
            t = environ.bool_var()
            f = environ.bool_var()
            d = environ.bool_var(True)

        cfg = environ.to_config(Cfg, environ={"APP_T": val, "APP_F": "nope"})

        assert cfg.t is True
        assert cfg.f is False
        assert cfg.d is True

    def test_tolerates_attribs(self):
        """
        Classes are allowed to have plain attr.ibs for e.g.
        __attrs_post_init__.
        """

        @environ.config
        class Cfg:
            e = environ.var()
            x = attr.ib(default=42)

        cfg = environ.to_config(Cfg, environ={"APP_E": "e"})

        assert Cfg("e", 42) == cfg

    def test_generate_help_str(self):
        """
        A help string is generated for a config class.

        Presence of defaults are indicated but they are not shown.
        """
        help_str = environ.generate_help(Parent)
        assert (
            help_str
            == """FOO_VAR1 (Required): var1, no default
FOO_VAR2 (Optional): var2, has default
FOO_VAR3 (Required): var3, bool_var, no default
FOO_VAR4 (Optional): var4, bool_var, has default
DOG (Optional): var5, named, has default
CAT (Required): var6, named, no default
FOO_CHILD_VAR7 (Required): var7, no default
FOO_CHILD_VAR8 (Optional): var8, has default
FOO_CHILD_VAR9 (Required): var9, bool_var, no default
FOO_CHILD_VAR10 (Optional): var10, bool_var, has default
DOG2 (Optional): var11, named, has default
CAT2 (Required): var12, named, no default
FOO_CHILD_VAR13 (Optional)
FOO_CHILD_VAR14 (Required)"""
        )

    def test_generate_help_str_with_defaults(self):
        """
        A help string is generated for a config class.

        display_defaults=True makes the defaults be shown.
        """
        help_str = environ.generate_help(Parent, display_defaults=True)

        assert (
            help_str
            == """FOO_VAR1 (Required): var1, no default
FOO_VAR2 (Optional, Default=bar): var2, has default
FOO_VAR3 (Required): var3, bool_var, no default
FOO_VAR4 (Optional, Default=True): var4, bool_var, has default
DOG (Optional, Default=canine): var5, named, has default
CAT (Required): var6, named, no default
FOO_CHILD_VAR7 (Required): var7, no default
FOO_CHILD_VAR8 (Optional, Default=bar): var8, has default
FOO_CHILD_VAR9 (Required): var9, bool_var, no default
FOO_CHILD_VAR10 (Optional, Default=True): var10, bool_var, has default
DOG2 (Optional, Default=canine): var11, named, has default
CAT2 (Required): var12, named, no default
FOO_CHILD_VAR13 (Optional, Default=default)
FOO_CHILD_VAR14 (Required)"""
        )

    def test_generate_help_str_when_prefix_is_empty(self):
        """
        Environment variables' names don't start with an underscore
        """
        help_str = environ.generate_help(NoPrefix)

        assert (
            help_str
            == """A_VAR (Required): a_var, no default
ANOTHER_VAR (Optional): another_var, has default
_START_WITH_UNDERSCORE (Required): this starts with an underscore"""
        )

    def test_custom_formatter(self):
        """
        Custom formatters can be passed and are used.
        """

        def bad_formatter(help_dicts):
            return "Not a good formatter"

        help_str = environ.generate_help(Parent, formatter=bad_formatter)

        assert help_str == "Not a good formatter"

    def test_frozen(self):
        """
        Frozen configurations are immutable.
        """

        @environ.config(frozen=True)
        class Cfg:
            x = environ.var()
            y = environ.var()

        cfg = environ.to_config(Cfg, {"APP_X": "foo", "APP_Y": "bar"})

        with pytest.raises(FrozenInstanceError):
            cfg.x = "next_foo"

        assert cfg.x == "foo"

    def test_frozen_child(self):
        """
        Frozen child groups are immutable.
        """

        @environ.config
        class Cfg:
            @environ.config(frozen=True)
            class Sub:
                z = environ.var()

            x = environ.var()
            y = environ.var()
            sub = environ.group(Sub)

        cfg = environ.to_config(
            Cfg, {"APP_X": "foo", "APP_Y": "bar", "APP_SUB_Z": "baz"}
        )

        cfg.x = "next_foo"

        assert cfg.x == "next_foo"

        with pytest.raises(FrozenInstanceError):
            cfg.sub.z = "next_baz"

        assert cfg.sub.z == "baz"

    def test_required_group_required_child_missing(self):
        """
        Groups are required if any of their child elements are.
        """

        @environ.config(prefix="PARENT")
        class WithRequiredChild:
            @environ.config(prefix="CHILD")
            class Child:
                grandchild = environ.var()

            child = environ.group(Child)

        cfg = environ.to_config(
            WithRequiredChild, {"PARENT_CHILD_GRANDCHILD": "FOO"}
        )
        assert cfg.child.grandchild == "FOO"
        with pytest.raises(environ.MissingEnvValueError) as e:
            environ.to_config(WithRequiredChild, dict())
        assert ("PARENT_CHILD_GRANDCHILD",) == e.value.args

    def test_required_group_optional_child_missing(self):
        """
        Required groups are fully structured if the only child is optional.
        """

        @environ.config(prefix="PARENT")
        class WithRequiredChild:
            @environ.config(prefix="CHILD")
            class Child:
                grandchild = environ.var("FOO")

            child = environ.group(Child)

        cfg = environ.to_config(WithRequiredChild, dict())
        assert cfg.child.grandchild == "FOO"

    def test_optional_group_required_child_missing(self):
        """
        Optional groups are set to `None` if a required child is missing.
        """

        @environ.config(prefix="PARENT")
        class WithOptionalChild:
            @environ.config(prefix="CHILD")
            class Child:
                grandchild = environ.var()

            child = environ.group(Child, optional=True)

        cfg = environ.to_config(WithOptionalChild, dict())
        assert cfg.child is None

    def test_optional_group_optional_child_missing(self):
        """
        Optional groups with an optional child is fully structured.
        """

        @environ.config(prefix="PARENT")
        class WithOptionalChild:
            @environ.config(prefix="CHILD")
            class Child:
                grandchild = environ.var("FOO")

            child = environ.group(Child, optional=True)

        cfg = environ.to_config(WithOptionalChild, dict())
        assert cfg.child.grandchild == "FOO"

    def test_optional_group_mixed_children_all_missing(self):
        """
        Optional groups are set to `None` if all mixed children are missing.
        """

        @environ.config(prefix="PARENT")
        class WithOptionalChild:
            @environ.config(prefix="CHILD")
            class Child:
                grandchild_a = environ.var()
                grandchild_b = environ.var("FOO")

            child = environ.group(Child, optional=True)

        cfg = environ.to_config(WithOptionalChild, dict())
        assert cfg.child is None

    def test_optional_group_mixed_children_optional_present(self):
        """
        Optional groups are required if any optional child is present.
        """

        @environ.config(prefix="PARENT")
        class WithOptionalChild:
            @environ.config(prefix="CHILD")
            class Child:
                grandchild_a = environ.var()
                grandchild_b = environ.var("FOO")

            child = environ.group(Child, optional=True)

        with pytest.raises(environ.MissingEnvValueError) as e:
            environ.to_config(
                WithOptionalChild, {"PARENT_CHILD_GRANDCHILD_B": "BAR"}
            )
        assert ("PARENT_CHILD_GRANDCHILD_A",) == e.value.args

    def test_nested_optional_group(self):
        """
        Optional groups which default to `None` count as required values.
        """

        @environ.config(prefix="PARENT")
        class WithOptionalGrandChild:
            @environ.config(prefix="CHILD")
            class Child:
                @environ.config(prefix="GRANDCHILD")
                class GrandChild:
                    foo = environ.var()

                grandchild = environ.group(GrandChild, optional=True)

            child = environ.group(Child)

        # By providing nothing, we expect the grandchild to default to `None`
        # which will fulfill the requirement that it have a value in the child
        cfg = environ.to_config(WithOptionalGrandChild, dict())
        assert cfg.child.grandchild is None
        # We also ensure that a properly set value is stored as expected
        cfg = environ.to_config(
            WithOptionalGrandChild, {"PARENT_CHILD_GRANDCHILD_FOO": "BAR"}
        )
        assert cfg.child.grandchild.foo == "BAR"



# Copyright 2021 Chris Rose

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import uuid

from unittest.mock import patch

import attr
import boto3
import pytest

from moto import mock_secretsmanager

import environ

from environ.exceptions import MissingSecretError
from environ.secrets import SecretsManagerSecrets, _SecretStr


@pytest.fixture(name="shut_boto_up", autouse=True, scope="session")
def _shut_boto_up():
    import logging

    for name in logging.Logger.manager.loggerDict.keys():
        if ("boto" in name) or ("urllib3" in name):
            logging.getLogger(name).setLevel(logging.WARNING)


@pytest.fixture(name="force_region", autouse=True, scope="session")
def _force_region():
    with patch.dict(os.environ, {"AWS_DEFAULT_REGION": "us-east-1"}):
        yield


@pytest.fixture(name="mock_aws_credentials")
def _mock_aws_credentials(force_region):
    with patch.dict(
        os.environ,
        {
            "AWS_ACCESS_KEY_ID": "testing",
            "AWS_SECRET_ACCESS_KEY": "testing",
            "AWS_SECURITY_TOKEN": "testing",
            "AWS_SESSION_TOKEN": "testing",
        },
    ):
        yield


@pytest.fixture(name="secretsmanager")
def _secretsmanager():
    with mock_secretsmanager():
        yield boto3.client("secretsmanager", region_name="us-east-2")


@pytest.fixture(name="secret")
def _secret():
    return str(uuid.uuid4())


@pytest.fixture(name="sm")
def _sm(secretsmanager, secret):
    secretsmanager.create_secret(Name=secret)
    secretsmanager.put_secret_value(SecretId=secret, SecretString="foobar")
    return SecretsManagerSecrets(client=secretsmanager)


class TestAWSSMSecret:
    def test_missing_default_raises(self, sm):
        """
        Missing values without a default raise an MissingSecretError.
        """

        @environ.config
        class Cfg:
            pw = sm.secret()

        with pytest.raises(MissingSecretError):
            environ.to_config(Cfg, {})

    def test_no_default_resolves_secret(self, sm, secretsmanager):
        """
        Verify that, when no default is provided, resolution should not happen
        until the configuration is needed.
        """

        @environ.config
        class Cfg:
            pw = sm.secret()

        secretsmanager.create_secret(Name="SecretName")
        secretsmanager.put_secret_value(
            SecretId="SecretName", SecretString="no-default"
        )
        assert (
            environ.to_config(Cfg, {"APP_PW": "SecretName"}).pw == "no-default"
        )

    def test_secret_works_with_default_client_overridden(
        self, mock_aws_credentials
    ):
        """
        Assert that the SM type can function without a custom client override
        when testing
        """
        sm = SecretsManagerSecrets()

        @environ.config
        class Cfg:
            pw = sm.secret()

        with mock_secretsmanager():
            # we need to make sure we're using the same region. It doesn't
            # matter which -- moto _and_ boto will try figure it out from the
            # environment -- but it has to be the same.
            sm.client.create_secret(Name="SecretName")
            sm.client.put_secret_value(
                SecretId="SecretName", SecretString="no-default"
            )
            conf = environ.to_config(Cfg, {"APP_PW": "SecretName"})
            assert conf.pw == "no-default"

    def test_default(self, sm, secret):
        """
        Defaults are used iff the key is missing.
        """

        @environ.config
        class Cfg:
            password = sm.secret(default="not used")
            secret = sm.secret(default="used!")

        cfg = environ.to_config(Cfg, {"APP_PASSWORD": secret})

        assert Cfg("foobar", "used!") == cfg

    def test_default_factory(self, sm, secret):
        """
        Defaults are used iff the key is missing.
        """

        def getpass():
            return "a default"

        @environ.config
        class Cfg:
            password = sm.secret(default=attr.Factory(getpass))
            secret = sm.secret(default=attr.Factory(getpass))

        cfg = environ.to_config(Cfg, {"APP_PASSWORD": secret})

        assert Cfg("foobar", "a default") == cfg

    def test_name_overwrite(self, sm, secret):
        """
        Passsing a specific key name is respected.
        """
        sm.client.put_secret_value(SecretId=secret, SecretString="foobar")

        @environ.config
        class Cfg:
            pw = sm.secret(name="password")

        cfg = environ.to_config(Cfg, {"password": secret})

        assert _SecretStr("foobar") == cfg.pw

    def test_nested(self, sm, secret):
        """
        Prefix building works.
        """
        sm.client.put_secret_value(SecretId=secret, SecretString="nested!")

        @environ.config
        class Cfg:
            @environ.config
            class DB:
                password = sm.secret()

            db = environ.group(DB)

        cfg = environ.to_config(Cfg, {"APP_DB_PASSWORD": secret})

        assert _SecretStr("nested!") == cfg.db.password



# Copyright 2017 Hynek Schlawack

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import secrets
from ._environ_config import (
    bool_var,
    config,
    generate_help,
    group,
    to_config,
    var,
)
from .exceptions import MissingEnvValueError


__version__ = "22.1.0"

__title__ = "environ_config"
__description__ = "Boilerplate-free configuration with env variables."
__uri__ = "https://github.com/hynek/environ_config"

__author__ = "Hynek Schlawack"
__email__ = "hs@ox.cx"

__license__ = "Apache-2.0"
__copyright__ = "Copyright (c) 2017 " + __author__


__all__ = [
    "MissingEnvValueError",
    "bool_var",
    "config",
    "generate_help",
    "group",
    "secrets",
    "to_config",
    "var",
]



# Copyright 2017 Hynek Schlawack

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import attr
import pytest

import environ

from environ.exceptions import MissingSecretError
from environ.secrets import (
    DirectorySecrets,
    INISecrets,
    VaultEnvSecrets,
    _SecretStr,
)


class TestSecretStr:
    def test_secret_str_no_repr(self):
        """
        Outside of reprs, _SecretStr behaves normally.
        """
        s = _SecretStr("abc")

        assert "'abc'" == repr(s)

    def test_secret_str_censors(self):
        """
        _SecretStr censors its __repr__ if it's called from another __repr__.
        """
        s = _SecretStr("abc")

        @attr.s
        class Cfg:
            s = attr.ib()

        assert "Cfg(s=<SECRET>)" == repr(Cfg(s))


@pytest.fixture
def ini_file(tmpdir):
    f = tmpdir.join("foo.ini")
    f.write(
        """\
[secrets]
password = foobar
db_password = nested!
[other_secrets]
password = bar%foo
[yet_another_section]
secret = qux
"""
    )
    return f


@pytest.fixture
def ini(ini_file):
    return INISecrets.from_path(str(ini_file))


class TestIniSecret:
    def test_missing_default_raises(self, ini):
        """
        Missing values without a default raise an MissingSecretError.
        """

        @environ.config
        class Cfg:
            pw = ini.secret()

        with pytest.raises(MissingSecretError):
            environ.to_config(Cfg, {})

    def test_default(self, ini):
        """
        Defaults are used iff the key is missing.
        """

        @environ.config
        class Cfg:
            password = ini.secret(default="not used")
            secret = ini.secret(default="used!")

        cfg = environ.to_config(Cfg, {})

        assert Cfg("foobar", "used!") == cfg

    def test_default_factory(self, ini):
        """
        Defaults are used iff the key is missing.
        """

        def getpass():
            return "thesecret"

        @environ.config
        class Cfg:
            password = ini.secret(default=attr.Factory(getpass))
            secret = ini.secret(default=attr.Factory(getpass))

        cfg = environ.to_config(Cfg, {})

        assert Cfg("foobar", "thesecret") == cfg

    def test_name_overwrite(self, ini):
        """
        Passsing a specific key name is respected.
        """

        @environ.config
        class Cfg:
            pw = ini.secret(name="password")

        cfg = environ.to_config(Cfg, {})

        assert _SecretStr("foobar") == cfg.pw

    def test_overwrite_sections(self, ini):
        """
        The source section can be overwritten.
        """
        ini.section = "yet_another_section"

        @environ.config
        class Cfg:
            password = ini.secret(section="other_secrets")
            secret = ini.secret()

        cfg = environ.to_config(Cfg, {})

        assert _SecretStr("bar%foo") == cfg.password

    def test_nested(self, ini):
        """
        Prefix building works.
        """

        @environ.config
        class Cfg:
            @environ.config
            class DB:
                password = ini.secret()

            db = environ.group(DB)

        cfg = environ.to_config(Cfg, {})

        assert _SecretStr("nested!") == cfg.db.password

    def test_from_path_in_env_delayed(self, ini_file):
        """
        `from_path_in_env` prepares for loading but doesn't load until
        `to_config` runs.
        """
        secret = INISecrets.from_path_in_env("APP_SECRETS_INI").secret

        @environ.config
        class Cfg:
            password = secret()

        cfg = environ.to_config(Cfg, {"APP_SECRETS_INI": str(ini_file)})

        assert "foobar" == cfg.password


@pytest.fixture
def vault():
    return VaultEnvSecrets(vault_prefix="SECRET")


class TestVaultEnvSecrets:
    def test_returns_secret_str(self, vault):
        """
        The returned strings are `_SecretStr`.
        """

        @environ.config
        class Cfg:
            x = vault.secret()

        cfg = environ.to_config(Cfg, {"SECRET_X": "foo"})

        assert isinstance(cfg.x, _SecretStr)
        assert "foo" == cfg.x

    def test_overwrite_name(self, vault):
        """
        The variable name can be overwritten.
        """

        @environ.config
        class Cfg:
            password = vault.secret(name="not_password")

        cfg = environ.to_config(
            Cfg, {"SECRET_PASSWORD": "wrong", "not_password": "correct"}
        )

        assert "correct" == cfg.password

    def test_missing_raises_missing_secret(self, vault):
        """
        Missing values without a default raise an MissingSecretError.
        """

        @environ.config
        class Cfg:
            pw = vault.secret()

        with pytest.raises(MissingSecretError):
            environ.to_config(Cfg, {})

    def test_prefix_callable(self):
        """
        vault_prefix can also be a callable that is called on each entry.
        """
        fake_environ = {"ABC_PW": "foo"}

        def extract(env):
            assert env == fake_environ
            return "ABC"

        vault = VaultEnvSecrets(vault_prefix=extract)

        @environ.config
        class Cfg:
            pw = vault.secret()

        cfg = environ.to_config(Cfg, fake_environ)

        assert _SecretStr("foo") == cfg.pw


@pytest.fixture
def secrets_dir(tmpdir):
    def make_secrets_file(name, content):
        secret_file = tmpdir.join(name)
        secret_file.write(content)

    make_secrets_file("empty", "")
    make_secrets_file("apples", "apples\n")
    make_secrets_file("oranges", "oranges")
    return str(tmpdir)


class TestDirectorySecrets:
    def test_secret_content(self, secrets_dir):
        """
        Reading secrets with different content.
        """
        dir = DirectorySecrets.from_path(secrets_dir)

        @environ.config
        class Cfg:
            empty = dir.secret()
            apples = dir.secret()
            oranges = dir.secret()

        cfg = environ.to_config(Cfg, {})

        assert _SecretStr("") == cfg.empty
        assert _SecretStr("apples\n") == cfg.apples
        assert _SecretStr("oranges") == cfg.oranges

    def test_secret_different_name(self, secrets_dir):
        """
        Attribute names can be different than filenames,
        the name argument can be used to select the filename.
        """
        dir = DirectorySecrets.from_path(secrets_dir)

        @environ.config
        class Cfg:
            nothing = dir.secret(name="empty")
            has_newline = dir.secret(name="apples")
            no_newline = dir.secret(name="oranges")

        cfg = environ.to_config(Cfg, {})

        assert _SecretStr("") == cfg.nothing
        assert _SecretStr("apples\n") == cfg.has_newline
        assert _SecretStr("oranges") == cfg.no_newline

    def test_secret_from_env(self, secrets_dir):
        """
        A directory can be specified in an environment variable.
        """
        dir = DirectorySecrets.from_path_in_env("SECRETS_DIR", "/tmp")

        @environ.config
        class Cfg:
            apples = dir.secret()

        cfg = environ.to_config(Cfg, {"SECRETS_DIR": secrets_dir})
        assert _SecretStr("apples\n") == cfg.apples

    def test_missing_default_value(self, tmpdir):
        """
        Missing secrets should raise Exceptions when loading the config.
        """
        dir = DirectorySecrets.from_path(str(tmpdir))

        @environ.config
        class Cfg:
            doesnt_exist = dir.secret()

        with pytest.raises(MissingSecretError):
            environ.to_config(Cfg, {})

    def test_default_value_specified(self, tmpdir):
        """
        Default values work for DirectorySecrets too.
        """
        dir = DirectorySecrets.from_path(str(tmpdir))

        @environ.config
        class Cfg:
            doesnt_exist = dir.secret(default="Test default value")

        cfg = environ.to_config(Cfg, {})
        assert "Test default value" == cfg.doesnt_exist



import attr

import environ


@environ.config(prefix="APP")
class AppConfig:
    host = environ.var("127.0.0.1")
    port = environ.var(5000, converter=int)


@environ.config(prefix="APP", from_environ="from_env")
class ConfigRenamed:
    host = environ.var("127.0.0.1", help="host help")
    port = environ.var(5000, converter=int, help="port help")


@environ.config(prefix="APP", from_environ="")
class ConfigEmptyName:
    host = environ.var("127.0.0.1", help="host help")
    port = environ.var(5000, converter=int, help="port help")


@environ.config(prefix="APP", from_environ=None)
class ConfigNoneName:
    host = environ.var("127.0.0.1", help="host help")
    port = environ.var(5000, converter=int, help="port help")


def test_has_classmethod():
    """
    Class based `from_environ` classmethod exists
    """
    assert hasattr(AppConfig, "from_environ")
    assert hasattr(ConfigRenamed, "from_env")
    assert not hasattr(ConfigEmptyName, "from_environ")
    assert not hasattr(ConfigNoneName, "from_environ")


def test_default():
    """
    Class based `from_environ` without `environ` argument.
    """
    cfg = AppConfig.from_environ()

    assert cfg.host == "127.0.0.1"
    assert cfg.port == 5000

    assert environ.to_config(AppConfig) == AppConfig.from_environ()
    assert environ.to_config(ConfigRenamed) == ConfigRenamed.from_env()


def test_env():
    """
    Class based `from_environ`  with explicit `environ` argument.
    """
    env = {"APP_HOST": "0.0.0.0"}
    cfg = AppConfig.from_environ(environ=env)

    assert cfg.host == "0.0.0.0"
    assert cfg.port == 5000

    assert environ.to_config(AppConfig, environ=env) == AppConfig.from_environ(
        environ=env
    )

    assert environ.to_config(
        ConfigRenamed, environ=env
    ) == ConfigRenamed.from_env(environ=env)


def test_factory_default():
    """
    Class based ``from_environ`` allows ``attr.Factory`` defaults.
    """

    @environ.config()
    class FactoryConfig:
        x = environ.var(attr.Factory(list))
        y = environ.var("bar")

    cfg = FactoryConfig.from_environ({"APP_Y": "baz"})

    assert cfg.x == []
    assert cfg.y == "baz"



import codecs

import attr

from environ._environ_config import Raise
from environ.exceptions import MissingSecretError


def _get_default_secret(var, default):
    """
    Get default or raise MissingSecretError.
    """
    if isinstance(default, attr.Factory):
        return attr.NOTHING
    elif isinstance(default, Raise):
        raise MissingSecretError(var)
    return default


def _open_file(path):
    return codecs.open(path, mode="r", encoding="utf-8")



# Copyright 2017 Hynek Schlawack

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Handling of sensitive data.
"""


import logging
import os
import sys

from configparser import NoOptionError, RawConfigParser

import attr

from environ._environ_config import CNF_KEY, RAISE, _ConfigEntry
from environ.exceptions import MissingSecretImplementationError

from ._utils import _get_default_secret, _open_file


try:
    from environ.secrets.awssm import SecretsManagerSecrets
except ImportError:  # pragma: no cover

    class SecretsManagerSecrets:
        def secret(self, *args, **kwargs):
            raise MissingSecretImplementationError(
                "AWS secrets manager requires boto3"
            )


log = logging.getLogger(__name__)


FileOpenError = OSError


@attr.s
class INISecrets:
    """
    Load secrets from an `INI file <https://en.wikipedia.org/wiki/INI_file>`_
    using `configparser.RawConfigParser`.
    """

    section = attr.ib()
    _cfg = attr.ib(default=None)
    _env_name = attr.ib(default=None)
    _env_default = attr.ib(default=None)

    @classmethod
    def from_path(cls, path, section="secrets"):
        """
        Look for secrets in *section* of *path*.

        :param str path: A path to an INI file.
        :param str section: The section in the INI file to read the secrets
            from.
        """
        return cls(section, _load_ini(path), None, None)

    @classmethod
    def from_path_in_env(cls, env_name, default=None, section="secrets"):
        """
        Get the path from the environment variable *env_name* **at runtime**
        and then load the secrets from it.

        This allows you to overwrite the path to the secrets file in
        development.

        :param str env_name: Environment variable that is used to determine the
            path of the secrets file.
        :param str default: The default path to load from.
        :param str section: The section in the INI file to read the secrets
            from.
        """
        return cls(section, None, env_name, default)

    def secret(
        self, default=RAISE, converter=None, name=None, section=None, help=None
    ):
        """
        Declare a secret on an `environ.config`-decorated class.

        :param str section: Overwrite the section where to look for the values.

        Other parameters work just like in `environ.var`.
        """
        if section is None:
            section = self.section

        return attr.ib(
            default=default,
            metadata={
                CNF_KEY: _ConfigEntry(name, default, None, self._get, help),
                CNF_INI_SECRET_KEY: _INIConfig(section),
            },
            converter=converter,
        )

    def _get(self, environ, metadata, prefix, name):
        # Delayed loading.
        if self._cfg is None and self._env_name is not None:
            log.debug("looking for env var '%s'.", self._env_name)
            self._cfg = _load_ini(
                environ.get(self._env_name, self._env_default)
            )

        ce = metadata[CNF_KEY]
        ic = metadata[CNF_INI_SECRET_KEY]
        section = ic.section

        if ce.name is not None:
            var = ce.name
        else:
            var = "_".join(prefix[1:] + (name,))
        try:
            log.debug("looking for '%s' in section '%s'.", var, section)
            val = self._cfg.get(section, var)
            return _SecretStr(val)
        except NoOptionError:
            return _get_default_secret(var, ce.default)


@attr.s
class DirectorySecrets:
    """
    Load secrets from a directory containing secrets in separate files.
    Suitable for reading Docker or Kubernetes secrets
    from the filesystem inside a container.

    .. versionadded:: 21.1.0
    """

    secrets_dir = attr.ib()
    _env_name = attr.ib(default=None)

    @classmethod
    def from_path(cls, path):
        """
        Look for secrets in *path* directory.

        :param str path: A path to directory containing secrets as files.
        """
        return cls(path)

    @classmethod
    def from_path_in_env(cls, env_name, default):
        """
        Get the path from the environment variable *env_name* and
        then load the secrets from that directory at runtime.

        This allows you to overwrite the path to the secrets directory
        in development.

        :param str env_name: Environment variable that is used to determine the
            path of the secrets directory.
        :param str default: The default path to load from.
        """
        return cls(default, env_name)

    def secret(self, default=RAISE, converter=None, name=None, help=None):
        return attr.ib(
            default=default,
            metadata={
                CNF_KEY: _ConfigEntry(name, default, None, self._get, help)
            },
            converter=converter,
        )

    def _get(self, environ, metadata, prefix, name):
        ce = metadata[CNF_KEY]
        # conventions for file naming might be different
        # than for environment variables, so we don't call .upper()
        filename = ce.name or "_".join(prefix[1:] + (name,))

        secrets_dir = environ.get(self._env_name, self.secrets_dir)
        secret_path = os.path.join(secrets_dir, filename)
        log.debug("looking for secret in file '%s'.", secret_path)

        try:
            with _open_file(secret_path) as f:
                val = f.read()
            return _SecretStr(val)
        except FileOpenError:
            return _get_default_secret(filename, ce.default)


@attr.s
class VaultEnvSecrets:
    """
    Loads secrets from environment variables that follow the naming style from
    `envconsul <https://github.com/hashicorp/envconsul>`_.
    """

    vault_prefix = attr.ib()

    def secret(self, default=RAISE, converter=None, name=None, help=None):
        """
        Almost identical to `environ.var` except that it takes *envconsul*
        naming into account.
        """
        return attr.ib(
            default=default,
            metadata={
                CNF_KEY: _ConfigEntry(name, default, None, self._get, help)
            },
            converter=converter,
        )

    def _get(self, environ, metadata, prefix, name):
        ce = metadata[CNF_KEY]

        if ce.name is not None:
            var = ce.name
        else:
            if callable(self.vault_prefix):
                vp = self.vault_prefix(environ)
            else:
                vp = self.vault_prefix
            var = "_".join((vp,) + prefix[1:] + (name,)).upper()

        log.debug("looking for env var '%s'.", var)
        try:
            val = environ[var]
            return _SecretStr(val)
        except KeyError:
            return _get_default_secret(var, ce.default)


class _SecretStr(str):
    """
    String that censors its __repr__ if called from an attrs repr.
    """

    def __repr__(self):
        # The frame numbers varies across attrs versions. Use this convoluted
        # form to make the call lazy.
        if (
            sys._getframe(1).f_code.co_name == "__repr__"
            or sys._getframe(2).f_code.co_name == "__repr__"
        ):
            return "<SECRET>"
        else:
            return str.__repr__(self)


CNF_INI_SECRET_KEY = CNF_KEY + "_ini_secret"


@attr.s
class _INIConfig:
    section = attr.ib()


def _load_ini(path):
    """
    Load an INI file from *path*.
    """
    cfg = RawConfigParser()
    with _open_file(path) as f:
        cfg.read_file(f)

    return cfg



import pytest

import environ


@environ.config(prefix="APP")
class AppConfig:
    host = environ.var("127.0.0.1", help="host help")
    port = environ.var(5000, converter=int, help="port help")


@environ.config(prefix="APP", generate_help="gen_help")
class ConfigRenamed:
    host = environ.var("127.0.0.1", help="host help")
    port = environ.var(5000, converter=int, help="port help")


@environ.config(prefix="APP", generate_help="")
class ConfigEmptyName:
    host = environ.var("127.0.0.1", help="host help")
    port = environ.var(5000, converter=int, help="port help")


@environ.config(prefix="APP", generate_help=None)
class ConfigNoneName:
    host = environ.var("127.0.0.1", help="host help")
    port = environ.var(5000, converter=int, help="port help")


def test_has_classmethod():
    """
    Class based `generate_help` classmethod exists
    """
    sentinel = object()
    # getattr returning default sentinel value means the att is missing
    # some attributes are expected
    assert getattr(AppConfig, "generate_help", sentinel) is not sentinel
    assert getattr(ConfigRenamed, "gen_help", sentinel) is not sentinel
    # another attributes shall be missing
    assert getattr(ConfigEmptyName, "generate_help", sentinel) is sentinel
    assert getattr(ConfigNoneName, "generate_help", sentinel) is sentinel


def test_generated_helps_equals():
    """
    Text by classmethod and environ.generate_help equals
    """
    assert environ.generate_help(AppConfig) == AppConfig.generate_help()
    assert environ.generate_help(ConfigRenamed) == ConfigRenamed.gen_help()


@pytest.mark.parametrize("display_defaults", [True, False])
def test_generated_helps_equals_display_defaults(display_defaults):
    """
    Text by classmethod and environ.generate_help equals with display_defaults
    """
    assert environ.generate_help(
        AppConfig, display_defaults=display_defaults
    ) == AppConfig.generate_help(display_defaults=display_defaults)
    assert environ.generate_help(
        ConfigRenamed, display_defaults=display_defaults
    ) == ConfigRenamed.gen_help(display_defaults=display_defaults)



# Copyright 2017 Hynek Schlawack

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import codecs
import os
import re

from setuptools import find_packages, setup


###############################################################################

NAME = "environ-config"
KEYWORDS = ["app", "config", "env", "cfg"]
PROJECT_URLS = {
    "Changelog": "https://environ-config.rtfd.io/en/stable/changelog.html",
    "Bug Tracker": "https://github.com/hynek/environ-config/issues",
    "Source Code": "https://github.com/hynek/environ-config",
    "Funding": "https://github.com/sponsors/hynek",
    "Ko-fi": "https://ko-fi.com/the_hynek",
}

META_PATH = os.path.join("src", "environ", "__init__.py")
CLASSIFIERS = [
    "Development Status :: 5 - Production/Stable",
    "Intended Audience :: Developers",
    "Natural Language :: English",
    "License :: OSI Approved :: Apache Software License",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
    "Programming Language :: Python :: Implementation :: CPython",
    "Programming Language :: Python :: Implementation :: PyPy",
    "Topic :: Software Development :: Libraries :: Python Modules",
]
PYTHON_REQUIRES = ">=3.7"
INSTALL_REQUIRES = ["attrs>=17.4.0"]
EXTRAS_REQUIRE = {
    "tests": ["pytest", "coverage[toml]", "moto"],
    "docs": ["sphinx", "furo"],
    "aws": ["boto3"],
}
EXTRAS_REQUIRE["dev"] = (
    EXTRAS_REQUIRE["aws"]
    + EXTRAS_REQUIRE["tests"]
    + EXTRAS_REQUIRE["docs"]
    + ["pre-commit"]
)

###############################################################################

HERE = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    """
    Build an absolute path from *parts* and and return the contents of the
    resulting file.  Assume UTF-8 encoding.
    """
    with codecs.open(os.path.join(HERE, *parts), "rb", "utf-8") as f:
        return f.read()


try:
    PACKAGES
except NameError:
    PACKAGES = find_packages(where="src")
try:
    META_PATH
except NameError:
    META_PATH = os.path.join(
        HERE, "src", NAME.replace("-", "_"), "__init__.py"
    )
finally:
    META_FILE = read(META_PATH)


def find_meta(meta):
    """
    Extract __*meta*__ from META_FILE.
    """
    meta_match = re.search(
        rf"^__{meta}__ = ['\"]([^'\"]*)['\"]", META_FILE, re.M
    )
    if meta_match:
        return meta_match.group(1)
    raise RuntimeError(f"Unable to find __{meta}__ string.")


VERSION = find_meta("version")
URI = find_meta("uri")
LONG = (
    read("README.rst")
    + "\n\n"
    + "Release Information\n"
    + "===================\n\n"
    + re.search(
        r"(\d+.\d.\d \(.*?\)\r?\n.*?)\r?\n\r?\n\r?\n----\r?\n\r?\n\r?\n",
        read("CHANGELOG.rst"),
        re.S,
    ).group(1)
    + (
        "\n\n`Full changelog <"
        "https://github.com/hynek/environ-config/blob/main/CHANGELOG.rst"
        ">`_.\n\n"
    )
    + read("AUTHORS.rst")
)


if __name__ == "__main__":
    setup(
        name=NAME,
        description=find_meta("description"),
        license=find_meta("license"),
        url=URI,
        project_urls=PROJECT_URLS,
        version=VERSION,
        author=find_meta("author"),
        author_email=find_meta("email"),
        maintainer=find_meta("author"),
        maintainer_email=find_meta("email"),
        keywords=KEYWORDS,
        long_description=LONG,
        long_description_content_type="text/x-rst",
        packages=PACKAGES,
        package_dir={"": "src"},
        zip_safe=False,
        classifiers=CLASSIFIERS,
        python_requires=PYTHON_REQUIRES,
        install_requires=INSTALL_REQUIRES,
        extras_require=EXTRAS_REQUIRE,
        options={"bdist_wheel": {"universal": "1"}},
    )



# Copyright 2021 Chris Rose

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Handling of sensitive data stored in AWS SecretsManager
"""
import logging

import attr
import boto3

from environ._environ_config import CNF_KEY, RAISE, _ConfigEntry

from ._utils import _get_default_secret


log = logging.getLogger(__name__)


def convert_secret(key):
    def converter(value):
        if isinstance(value, str):
            return value
        return value[key]

    return converter


def _build_secretsmanager_client():
    client = boto3.client("secretsmanager")
    log.debug("Created a secretsmanager client %s", client)
    return client


@attr.s(init=False)
class SecretsManagerSecrets:
    """
    Load secrets from the AWS secretsmanager.

    The secret name should be stored in the environment variable

    .. versionadded:: 21.4.0
    """

    def __init__(self, client=None):
        self._client = client

    @property
    def client(self):
        if self._client is None:
            self._client = _build_secretsmanager_client()
        return self._client

    def secret(
        self,
        default=RAISE,
        converter=convert_secret("SecretString"),
        name=None,
        help=None,
    ):
        """
        Declare a secrets manager secret on an `environ.config`-decorated class

        All parameters work just like in `environ.var`.
        """
        return attr.ib(
            default=default,
            metadata={
                CNF_KEY: _ConfigEntry(name, default, None, self._get, help)
            },
            converter=converter,
        )

    def _get(self, environ, metadata, prefix, name):
        ce = metadata[CNF_KEY]
        if ce.name:
            secret_name_envvar = ce.name
            log.debug(
                "override env variable with explicit name %s",
                secret_name_envvar,
            )
        else:
            parts = prefix + (name,)
            secret_name_envvar = "_".join(parts).upper()
            log.debug(
                "secret name environment variable %s", secret_name_envvar
            )

        try:
            secret_name = environ[secret_name_envvar]
        except KeyError:
            # missing the environment; let's try to get the default
            log.debug(
                "no key %s in environment, using default=%s",
                secret_name_envvar,
                ce.default,
            )
            return _get_default_secret(secret_name_envvar, ce.default)
        log.debug("secret name: %s", secret_name)

        return self.client.get_secret_value(SecretId=secret_name)
